################################################################################
# Predicting the cell type from methylation data
#-------------------------------------------------------------------------------
# author: Fabian Müller <fmueller@mpi-inf.mpg.de>
# date:   2016-09-27
#-------------------------------------------------------------------------------
# INPUT:
#   - Matrix of methylation levels (tsv)
#   - Sample annotation
#   - output directory
# OUTPUT:
#   - multiple plots
#   - trained cell type predictors (RDS files)
# Example call:
# Rscript cellTypePredictor.R --features methMatrix.tsv --annot sampleAnnot.tsv --out predictCellCellTypes
################################################################################
library(RnBeads) # logging and utilities
library(argparse) # command line parsing
library(matrixStats)

library(glmnet) # elastic net
library(e1071) # SVM
################################################################################
# Arguments and Parameters
################################################################################
ap <- ArgumentParser()
ap$add_argument("-f", "--features", action="store", dest="feats", help="Filename of the feature/methylation matrix (tsv, dimension: features X samples, includes a header).")
ap$add_argument("-a", "--annot", action="store", dest="annot", help="Filename of the sample annotation matrix (tsv). Must contain the same number of rows as the feature matrix has columns and a header line. The column called 'class' is assumed to contain the class assignment. An optional 'train' column may contain TRUE/FALSE depending on whether a sample will be used in training of the predictor. An optional 'color' column may contain color values for each class (colors must be consistent for samples of the same class).")
ap$add_argument("-o", "--out", action="store", dest="out", help="Output directory. Must be nonexisting and will be created.")
ap$add_argument("-p", "--predictor", action="store", dest="predResFile", default="", help="(Optional) Prediction result file (rds) as output by this script. If specified, no predictor is trained, but the supplied predictor is used to classify the specified dataset.")
cmdArgs <- ap$parse_args()

argv <- commandArgs(trailingOnly = FALSE)
################################################################################
# Preliminaries and globals
################################################################################
SCRIPT.DIR <- dirname(substring(argv[grep("--file=", argv)], 8)) # location of the this script
SRC.DIR <- file.path(SCRIPT.DIR, "src")
OUT.DIR <- cmdArgs$out

MISSING.DATA.METHOD <- "impute.knn5"
DO.CV <- TRUE # perform cross validation
DO.LOCO <- TRUE # fit leave-one-class-out classifiers

COLOR.PALETTE.CAT <- c("#2166AC","#B2182B","#1B9E77","#D95F02","#7570B3","#E7298A","#66A61E","#E6AB02","#A6761D","#666666","#00441B","#40004B","#053061")
################################################################################
# Load methods
################################################################################
source(file.path(SRC.DIR, "utils.R"))
source(file.path(SRC.DIR, "utils_prediction.R"))
################################################################################
# Main
################################################################################
# set ggplot theme
theme_set(theme_bw())

logger.start("Loading data")
	logger.status("Feature matrix")
	if (!file.exists(cmdArgs$feats)) logger.error(c("Could not find feature matrix file @", cmdArgs$feats))
	mm <- as.matrix(read.table(cmdArgs$feats, sep="\t", header=TRUE, comment.char="", stringsAsFactors=FALSE))
	logger.info(c("Loaded", nrow(mm), "features for", ncol(mm), "samples"))

	logger.status("Annotation table")
	if (!file.exists(cmdArgs$annot)) logger.error(c("Could not find annotation table @", cmdArgs$annot))
	aa <- read.table(cmdArgs$annot, sep="\t", header=TRUE, comment.char="", stringsAsFactors=FALSE) 
	if (nrow(aa) != ncol(mm)) logger.error(c("Annotation file must have the same number of rows (", nrow(aa), ") as the feature matrix has columns (", ncol(mm), ")"))
	isTrain <- rep(TRUE, nrow(aa))
	if (is.element('train', colnames(aa))){
		if (!is.logical(aa[,"train"])) logger.error("The 'train' column must contain logical values")
		isTrain <- aa[,"train"]
		logger.info(c("Number of samples         :", length(isTrain)))
		logger.info(c("Number of training samples:", sum(isTrain)))
	}
	if (!is.element('class', colnames(aa))) logger.error("The annotation table does not contain a class column")
	classes <- getOrderedCellTypeFactor(aa[,"class"])
	classes.train <- droplevels(classes[isTrain])
	logger.info(c("Number of classes         :", length(levels(classes.train))))
	
	# set up color scheme
	sampleColors <- NULL
	colorScheme <- rep(COLOR.PALETTE.CAT, length.out=length(levels(classes)))
	if (is.element('color', colnames(aa))){
		sampleColors <- aa[,"color"]
		logger.status("Preparing color scheme")
		if (any(is.na(aa[isTrain,"color"]))) logger.error("invalid color column")
		colorScheme <- sapply(levels(classes), FUN=function(cc){
			curColor <- unique(aa[classes==cc, "color"])
			if (length(curColor)!=1) logger.error("could not parse color scheme")
			return(curColor)
		})
	}
	names(colorScheme) <- levels(classes)

	predRes <- NULL
	if (nchar(cmdArgs$predResFile) > 0){
		logger.status("Prediction result")
		if (!file.exists(cmdArgs$predResFile)) logger.error(c("Invalid prediction result file @", cmdArgs$predResFile))
		predRes <- readRDS(cmdArgs$predResFile)
	}
logger.completed()

logger.start("Preparing output directory")
	if (dir.exists(OUT.DIR)){
		logger.error(c("Output directory already exists @", OUT.DIR))
	} else {
		dir.create(OUT.DIR, recursive=TRUE)
		logger.info(c("Created output directory:",OUT.DIR))
	}
logger.completed()

logger.start("Preparing data")
	mm.nao <- mm
	if (any(is.na(mm))){
		missData <- handleMissingData(mm, MISSING.DATA.METHOD)
		mm.nao <- missData$data
	}
logger.completed()

do.train <- is.null(predRes)
if (do.train){
	# took about 1 hour for training based on 319 samples and 492102 features on our computing infrastructure
	logger.start("Training predictor(s)")
		predRes <- fitModels(mm.nao, classes, isTrain, "predict", doPredict=TRUE, doPlot=TRUE, colScheme=colorScheme, en.loocv=FALSE, models2fit=c("en", "svm"))
	logger.completed()

	logger.start("Exporting tables")
		pref <- file.path(OUT.DIR, "predictionTable")
		exportPredictorTables(predRes, pref)
	logger.completed()

	#large output (>3GB)
	logger.start("Saving prediction result")
		fn <- file.path(OUT.DIR, "predictionResult.rds")
		saveRDS(predRes, fn)
	logger.completed()

	if (DO.CV){
		logger.start("Fitting models using 10-fold Cross-Validation")
			predRes.cv <- performCrossVal(mm.nao[,isTrain], classes.train, "predict", doPlot=TRUE, colScheme=colorScheme)
		logger.completed()
	}
	if (DO.LOCO){
		logger.start("Leave-one-class-out (LOCO) classifiers")
			for (cl in levels(classes.train)){
				logger.start(c("Predicting", cl))
					isTrain.loco <- isTrain & classes!=cl
					predRes.loco <- fitModels(mm.nao, classes, isTrain.loco, paste0("predict_loco_", cl), doPredict=TRUE, doPlot=TRUE, colScheme=colorScheme)
					logger.status("Exporting tables")
					pref <- file.path(OUT.DIR, paste0("predictionTable_loco_", cl))
					exportPredictorTables(predRes.loco, pref)
					logger.status("Saving LOCO prediction result")
					fn <- file.path(OUT.DIR, paste0("predictionResult_loco_", cl, ".rds"))
					saveRDS(predRes.loco, fn)
				logger.completed()
			}
		logger.completed()
	}
} else {
	logger.start("Prediction using specified predictor(s)")
		nFeats.pred <- length(predRes[[1]]$fit$imp) #number of features used by the predictor(s)
		if (nrow(mm.nao) != nFeats.pred) logger.error(c("The number of features used by the predictor(s) (", nFeats.pred, ") does not match the number of features in the data matrix (", nrow(mm.nao), ")"))
		preds <- doPrediction(predRes, t(mm.nao), y=classes, sampleColors=sampleColors, fileSuffix="predict", doPlot=TRUE)
		pref <- file.path(OUT.DIR, "predictionTable")
		exportPredictorTables(preds, pref)
	logger.completed()
}
