################################################################################
# Utilities for predicting cell types based on DNA methylation data
#-------------------------------------------------------------------------------
# author: Fabian Müller
# date:   2016-09-13
################################################################################
# return a list of cross-validation sets containing indices
getCrossValInds <- function(N, k=10){
	inds <- 1:N
	indList <- list()
	N.fold <- ceiling(N/k)
	while (length(inds)>0){
		if (length(inds) > N.fold){
			curInds <- sort(sample(inds, N.fold))
		} else {
			curInds <- inds
		}
		indList <- c(indList, list(curInds))
		inds <- setdiff(inds, curInds)
	}
	return(indList)
}

# fit an elastic net general linear model for predicting y based on X
fitElasticNet <- function(X, y, ...){
	require(glmnet)
	multiclass <- length(unique(y)) > 2
	classImp <- NULL
	if (multiclass){
		fit.cv <- cv.glmnet(X, y, family="multinomial", type.multinomial="grouped", parallel=FALSE, ...)
		probFun <- function(mod, X) {predict(mod, X, type="response", s="lambda.1se")[,,1]}
		probs <- probFun(fit.cv, X)
		betas <- coef(fit.cv, s="lambda.1se")
		betas <- do.call("cbind", lapply(betas, FUN=function(x){x[,1]}))
		#aggregate coefficient/importance: L2 norm of coefficients for regions
		imps <- betas[-1,] #remove the intercept
		classImp <- imps
		imps <- rowSums(imps^2)^(1/2) #L2 Norm per feature
		imp.type <- "coef_l2norm"
	} else {
		fit.cv <- cv.glmnet(X, y, family="binomial", type.multinomial="grouped", parallel=FALSE, ...)
		probFun <- function(mod, X) {
			res <- predict(mod, X, type="response", s="lambda.1se")[,1]
			res <- cbind(1-res, res)
			colnames(res) <- mod$glmnet.fit$classnames
			return(res)
		}
		probs <- probFun(fit.cv, X)
		betas <- coef(fit.cv, s="lambda.1se")
		imps <- betas[-1,]
		imp.type <- "coef"
	}
	predFun <- function(mod, X) {predict(mod, X, type="class", s="lambda.1se")[,1]}
	preds <- predFun(fit.cv, X)
	is.imp <- abs(imps) > 0 #select important features

	res <- list(
		model=fit.cv,
		pred=preds,
		prob=probs,
		imp=imps,
		classImp=classImp,
		is.imp=is.imp,
		imp.type=imp.type,
		lamdba.sel=fit.cv$lambda.1se,
		pred.fun=predFun,
		prob.fun=probFun
	)
	return(res)
}

# fit a random forest model for predicting y based on X
fitRandomForest <- function(X, y, ...){
	require(randomForest)
	fit <- randomForest(X, y, ...)
	predFun <- function(mod, X) {predict(mod, X, type="response")}
	preds <- predFun(fit, X)
	probFun <- function(mod, X) {predict(mod, X, type="prob")}
	probs <- probFun(fit, X)
	imps <- fit$importance[,"MeanDecreaseGini"]
	is.imp <- imps > quantile(imps, probs=0.9) #important features: the 10% features with highest mean decrease in gini
	res <- list(
		model=fit,
		pred=preds,
		prob=probs,
		imp=imps,
		is.imp=is.imp,
		classImp=NULL,
		imp.type="MeanDecreaseGini",
		pred.fun=predFun,
		prob.fun=probFun
	)
	return(res)
}

# compute coefficients from a trained linear SVM model
svm.compute.coefficients <- function(mod){
	if (mod$nclasses < 3){
		alpha <- mod$coefs
		beta <- colSums(alpha[,1] * mod$SV) #multiply the alpha vector with the support vectors and sum; R: vector * matrix means multiply the vector with each column of the matrix
		beta.intercept <- -mod$rho
		beta <- c(beta.intercept, beta)
	} else {
		# multiclass case
		# code adapted from e1071 package : inst/doc/svminternals.pdf: 
		# https://stat.ethz.ch/pipermail/r-help/2002-January/017795.html
		start <- c(1, cumsum(mod$nSV)+1)
		start <- start[-length(start)]
		getBinaryBetas <- function(i, j){
			## ranges for class i and j:
			ri <- start[i] : (start[i] + mod$nSV[i] - 1)
			rj <- start[j] : (start[j] + mod$nSV[j] - 1)
			## coefs for (i,j):
			coef1 <- mod$coefs[ri, j-1]
			coef2 <- mod$coefs[rj, i]
			betas <- crossprod(coef1, mod$SV[ri,]) + crossprod(coef2, mod$SV[rj,])
			return(betas)
		}
		beta <- matrix(numeric(0),nrow=length(mod$rho), ncol=ncol(mod$SV)+1)
		colnames(beta) <- c("(Intercept)", colnames(mod$SV))
		beta[,1] <- -mod$rho
		counter <- 0
		for (i in 1 : (mod$nclasses - 1)) {
		    for (j in (i + 1) : mod$nclasses){
		    	counter <- counter + 1
		    	beta[counter, 2:ncol(beta)] <- getBinaryBetas(i, j)
		    }
		}
	}
	return(beta)
}
# fit a (linear) SVM model for predicting y based on X
fitSVM <- function(X, y, kernel="linear", ...){
	require(e1071)
	fit <- svm(x=X, y=y, kernel=kernel, probability=TRUE, ...)
	predFun <- function(mod, X){predict(mod, X)}
	preds <- predFun(fit, X)
	probFun <- function(mod, X){attr(predict(mod, X, probability=TRUE), "probabilities")}
	probs <- probFun(fit, X)
	beta <- svm.compute.coefficients(fit)
	classImp <- NULL
	if (is.matrix(beta)){ #multiclass
		zscores <- apply(beta[,-1], 1, FUN=function(x){abs(x-mean(x))/sd(x)})
		binLabels <- colnames(fit$decision.values) #binary classification labels
		colnames(zscores) <- binLabels
		imps <- rowMaxs(zscores)
		imp.type <- "SVM_linear_max_pairwise_zscore"
		#class impartance as the maximum importance of all pairwise comparisons involving that class
		classImp <- do.call("cbind", lapply(colnames(probs), FUN=function(cc){
			rowMaxs(zscores[, grepl(cc, binLabels), drop=FALSE])
		}))
		colnames(classImp) <- colnames(probs)
	} else { # binary classification
		bb <- beta[-1]
		zscores <- abs(bb-mean(bb))/sd(bb)
		imps <- zscores
		imp.type <- "SVM_linear_zscore"
	}
	is.imp <- imps > 1.96
	res <- list(
		model=fit,
		pred=preds,
		prob=probs,
		imp=imps,
		classImp=classImp,
		is.imp=is.imp,
		imp.type=imp.type,
		pred.fun=predFun,
		prob.fun=probFun
	)
	return(res)
}

# plot the class probabilities and predictions in a heatmap
plotClassProbHm <- function(preds, probs){
	df2p <- melt(probs)
	colnames(df2p) <- c("sample", "cellType", "classProb")
	df2p$sample <- factor(df2p$sample, levels=sort(levels(df2p$sample), decreasing=TRUE))
	#order the columns according to desired cell type ordering
	df2p$cellType <- getOrderedCellTypeFactor(df2p$cellType)
	df2p$prediction <- factor(preds[as.character(df2p$sample)], levels=levels(df2p$cellType))
	pp <- ggplot(df2p) + aes(x=cellType, y=sample, fill=classProb) + geom_tile() +
	  	  scale_fill_gradientn(colours=brewer.pal(9,"OrRd")) +
	  	  geom_point(aes(x=prediction,fill=NULL)) +
	  	  theme(axis.text.x = element_text(angle=90, hjust=1, vjust=0.5))
	return(pp)
}

# plot spider/radar plots given class probabilities and predictions
plotClassProbRadar <- function(preds, probs, fn, sampleColors=NULL){
	df2p <- melt(probs)
	colnames(df2p) <- c("sample", "cellType", "classProb")
	df2p$sample <- factor(df2p$sample, levels=sort(levels(df2p$sample), decreasing=TRUE))
	#order the columns according to desired cell type ordering
	df2p$cellType <- getOrderedCellTypeFactor(df2p$cellType)
	df2p$prediction <- factor(preds[as.character(df2p$sample)], levels=levels(df2p$cellType))
	samples <- sort(as.character(unique(df2p$sample)))
	pps <- lapply(samples, FUN=function(ss){
		df2p.cur <- df2p[df2p$sample==ss,]
		color.cur <- "black"
		if (!is.null(sampleColors) && is.element(ss, names(sampleColors))) color.cur <- sampleColors[ss]
		pp <- ggplot(df2p.cur) + aes(x=cellType, y=classProb) + geom_bar(stat = "identity", width=1, fill=color.cur) + 
			  scale_y_continuous(limits=c(0, 1)) + coord_polar() +
			  ggtitle(ss) + 
			  theme(plot.title=element_text(size=36, face="bold"), panel.border=element_blank(), axis.title.x=element_blank(), axis.ticks=element_blank())
		return(pp)
	})
	nplots <- length(samples)
	ncol <- 10
	pdf(fn, width=10*ncol, height=10*nplots/ncol)
		do.call("grid.arrange", c(pps, ncol=10))
	dev.off()
}

# plot a heatmap of class importances
plotClassImpHeatmap <- function(classImps, colScheme=NULL, color.panel=colorpanel(100, brewer.pal(7,"PiYG")[1], brewer.pal(7,"PiYG")[4], brewer.pal(7,"PiYG")[7]), ...){
	require(gplots)
	if (is.null(classImps)){
		logger.error("object does not contain class importances")
	} else {
		colColors <- FALSE
		curMat <- classImps
		if (!is.null(colScheme)) {
			#order the matrix and colors by cell type
			cts <- levels(getOrderedCellTypeFactor(colnames(curMat)))
			colColors <- colScheme[cts]
			curMat <- curMat[,cts]
		}
		heatmap.2(curMat, trace="none", scale="none", col=color.panel, ColSideColors=colColors, dendrogram="row", Colv=FALSE, labRow=FALSE, ...)
	}
}

# given a score and labels get the main ROC results for later plotting
getRocResult <- function(scores, labs){
	require(ROCR)
	isCv <- is.list(labs)
	pred <- prediction(scores, labs)
	curveObj <- performance(pred, 'tpr', 'fpr')
	aucObj <- performance(pred, 'auc')
	aucs <- unlist(aucObj@y.values)
	res <- list(
		curve=curveObj,
		auc=aucs
	)
	return(res)
}

# Plot ROC curves given class probabilities and true labels
plotROC <- function(probs, labs, fn, colScheme=NULL){
	isCv <- is.list(labs)
	classes <- sort(unique(as.character(unlist(labs))))
	#get ROC curve results per class
	getScore <- function(probs, class){
		iclass <- match(as.character(class), colnames(probs))
		if (length(iclass)!=1 || any(is.na(iclass))) stop(paste("invalid class:",class))
		return(probs[, iclass] - rowMaxs(probs[, -iclass, drop=FALSE], na.rm=TRUE))
	}
	rocRes <- lapply(classes, FUN=function(cc){
		if (isCv){
			scs <- lapply(probs, FUN=function(pp){getScore(pp, cc)})
			labs.bin <- lapply(labs, FUN=function(ll){as.numeric(ll==cc)})
			#remove folds where all binary labels are identical
			keepFolds <- sapply(labs.bin, FUN=function(x){length(unique(x))>1})
			scs <- scs[keepFolds]
			labs.bin <- labs.bin[keepFolds]
		} else {
			scs <- getScore(probs, cc)
			labs.bin <- as.numeric(labs==cc)
		}
		res <- getRocResult(scs, labs.bin)
		return(res)
	})
	names(rocRes) <- classes
	#plotting
	unitLength <- 5
	pdf(fn, height=unitLength, width=(length(classes)+1)*unitLength)
		par(mfrow=c(1,length(classes)+1))
		aucMeans <- c()
		for (cc in classes){
			curColor <- colScheme[cc]
			if(is.null(curColor)) curColor <- "black"
			rr <- rocRes[[cc]]
			meanAuc.cur <- mean(rr$auc)
			plot(rr$curve, avg='vertical', spread.estimate='stddev', lwd=5, main=cc, col=curColor)
			text(x=0.6, y=0, labels=paste0("Nfolds:",length(rr$auc), "; (mean) AUC: ", round(meanAuc.cur, 2)), adj=0)
			aucMeans <- c(aucMeans, meanAuc.cur)
		}
		# add summary to last panel
		plot.new()
		text(0.5, 0.5, labels=paste0("Mean mean AUC: ", round(mean(aucMeans), 2)))
	dev.off()
}

# plot a class confusion matrix given class probabilities and true labels
plotConfusionTable <- function(preds, labs, colorPanel=brewer.pal(9,"Blues"), normRows=FALSE){
	predFac <- getOrderedCellTypeFactor(preds)
	labFac <- getOrderedCellTypeFactor(labs)
	tt <- table(labFac, predFac)
	df2p <- melt(tt)
	colnames(df2p) <- c("label", "predicted", "count")
	df2p$predicted <- factor(df2p$predicted, levels=levels(predFac))
	df2p$label <- factor(df2p$label, levels=rev(levels(labFac))) #reverse labels for plotting vertically
	df2p$countForColor <- df2p$count
	if (normRows){
		for (ll in levels(df2p$label)){
			inds <- df2p$label == ll
			total <- sum(df2p$count[inds])
			df2p$countForColor[inds] <- df2p$count[inds]/total
		}
	}
	pp <- ggplot(df2p) + aes(x=predicted, y=label, fill=countForColor) + geom_tile() +
		  # coord_fixed() +
		  scale_fill_gradientn(colours=colorPanel)
	pp <- pp + geom_text(aes(label=count))
	pp <- pp + theme(
					axis.ticks = element_blank(),
				  	panel.grid.major = element_blank(),
				  	panel.background = element_blank(),
		  		    panel.border = element_blank(),
		  		    axis.text.x = element_text(angle = 90, hjust = 1)
			  	)
	if (normRows){
		pp <- pp + guides(fill=FALSE)
	}
	pp
}

# given a data structure containing trained models, perform prediction for new data
# for all models in the data structure
doPrediction <- function(predRes, X, y=NULL, sampleColors=NULL, fileSuffix="", doPlot=TRUE){
	res <- predRes
	for (cl in names(predRes)){
		logger.start(c("Predicting using model:", cl))
			probs <- predRes[[cl]]$fit$prob.fun(predRes[[cl]]$fit$model, X)
			preds <- predRes[[cl]]$fit$pred.fun(predRes[[cl]]$fit$model, X)
			if (length(rownames(X)) == length(preds)) names(preds) <- rownames(X)
			res[[cl]][["pred"]] <- preds
			res[[cl]][["prob"]] <- probs

			if (doPlot){
				pp <- plotClassProbHm(preds, probs)
				ggsave(file.path(OUT.DIR, paste0("hm_", cl, "_classProbs_",fileSuffix,".pdf")), pp, width=8, height=24)
				fn <-  file.path(OUT.DIR, paste0("rd_", cl, "_classProbs_",fileSuffix,".pdf"))
				plotClassProbRadar(preds, probs, fn, sampleColors)
				if (length(y)==nrow(X)){
					pp <- plotConfusionTable(preds, y)
					ggsave(file.path(OUT.DIR, paste0("confusion_", cl, "_",fileSuffix,".pdf")), pp, width=10, height=10)
				}
			}
		logger.completed()
	}
	return(res)
}

# fit a series of classification models based on a matrix of methylation levels (regions X samples)
# store them in a data structure for later reuse
fitModels <- function(methMat, ct, is.train=rep(TRUE, ncol(methMat)), fileSuffix="", doPredict=TRUE, doPlot=FALSE, colScheme=NULL, en.loocv=FALSE, models2fit=c("en", "svm")){
	if (!(is.integer(is.train) || is.logical(is.train))){
		logger.error("Invalid value for is.train")
	}
	if (is.logical(is.train) && sum(is.train) < 2){
		logger.error("insufficient training points")
	}
	y <- ct[is.train]
	if (is.factor(y)){
		y <- droplevels(y)
	} else {
		y <- factor(y)
	}
	X <- t(methMat[,is.train])
	sampleColors <- colScheme[as.character(ct)]
	if (!is.null(sampleColors) && !all(is.na(sampleColors))) names(sampleColors) <- colnames(methMat)

	res <- list()
	if (is.element("en", models2fit)) {
		logger.start("Fitting Elastic Net (en) model")
			if (en.loocv) {
				logger.info("Fitting model using Leave-One-Out Cross-Validation")
				fit.en <- fitElasticNet(X, y, alpha=0.5, nfolds=length(y))
			} else {
				logger.info("Fitting model using 10-fold Cross-Validation")
				fit.en <- fitElasticNet(X, y, alpha=0.5)
			}
			res[["en"]] <- list(fit=fit.en, pred=NULL, prob=NULL)
		logger.completed()
	}
	if (is.element("rf", models2fit)) {
		logger.start("Fitting Random Forest (rf) model")
			fit.rf <- fitRandomForest(X, y)
			res[["rf"]] <- list(fit=fit.rf, pred=NULL, prob=NULL)
		logger.completed()
	}
	if (is.element("svm", models2fit)) {
		logger.start("Fitting linear SVM model")
			fit.svm <- fitSVM(X, y)
			res[["svm"]] <- list(fit=fit.svm, pred=NULL, prob=NULL)
		logger.completed()
	}
	if (doPredict){
		logger.start("Prediction and plotting")
			res <- doPrediction(res, t(methMat), ct, sampleColors, fileSuffix, doPlot=doPlot)
		logger.completed()
	}
	return(res)
}

# fit a series of classification models based on a matrix of methylation levels (regions X samples)
# in a cross-validation setting
# store them in a data structure for later reuse
performCrossVal <- function(methMat, ct, fileSuffix="cv", doPlot=FALSE, colScheme=NULL, nfolds=10){
	classLabs <- sort(unique(as.character(ct)))
	if (is.factor(ct)) classLabs <- levels(droplevels(ct))
	sampleNames <- colnames(methMat)
	sampleColors <- colScheme[as.character(ct)]
	if (!is.null(sampleColors)) names(sampleColors) <- sampleNames
	cvInds <- getCrossValInds(length(ct), k=nfolds)
	en.loocv <- nfolds==length(ct)
	predTest.folds <- lapply(1:length(cvInds), FUN=function(i){
		logger.start(c("Fold:",i))
		fold <- cvInds[[i]]
		is.train <- setdiff(1:ncol(methMat), fold)
		obj <- fitModels(methMat, ct, is.train, fileSuffix, doPredict=TRUE, doPlot=FALSE, en.loocv=en.loocv)
		pred <- do.call("cbind", lapply(obj, FUN=function(x){as.character(x$pred[fold])}))
		prob <- lapply(obj, FUN=function(x){x$prob[fold,classLabs]})
		res <- list(obj=obj, pred=pred, prob=prob, labs=ct[fold])
		logger.completed()
		return(res)
	})
	classifiers <- colnames(predTest.folds[[1]]$pred)
	predTab <- matrix(character(0),nrow=length(ct),ncol=ncol(predTest.folds[[1]]$pred))
	rownames(predTab) <- sampleNames
	colnames(predTab) <- colnames(predTest.folds[[1]]$pred)
	probArr <- array(numeric(0), dim=c(length(ct), length(classLabs), length(predTest.folds[[1]]$prob)), dimnames=list(sampleNames, classLabs, names(predTest.folds[[1]]$prob)))
	for (i in 1:length(cvInds)){
		predTab[cvInds[[i]],] <- predTest.folds[[i]]$pred
		for (cl in classifiers){
			probArr[cvInds[[i]], , cl] <- predTest.folds[[i]]$prob[[cl]]
		}
	}
	misclass <- apply(predTab, 2, FUN=function(x){sum(x!=as.character(ct))})
	if (doPlot){
		labs.cv <- lapply(predTest.folds, FUN=function(x){x$labs})
		for (cl in classifiers){
			preds <- predTab[, cl]
			probs <- probArr[,,cl]
			pp <- plotClassProbHm(preds, probs)
			ggsave(file.path(OUT.DIR, paste0("hm_",cl,"_cv_classProbs_",fileSuffix,".pdf")), pp, width=8, height=24)
			fn <-  file.path(OUT.DIR, paste0("rd_",cl,"_cv_classProbs_",fileSuffix,".pdf"))
			plotClassProbRadar(preds, probs, fn, sampleColors)
			#confusion table
			pp <- plotConfusionTable(preds, ct)
			ggsave(file.path(OUT.DIR, paste0("confusion_",cl,"_cv_",fileSuffix,".pdf")), pp, width=10, height=10)
			#ROC curves
			fn <-  file.path(OUT.DIR, paste0("roc_",cl,"_cv_classProbs_",fileSuffix,".pdf"))
			probs.cv <- lapply(predTest.folds, FUN=function(x){x$prob[[cl]]})
			plotROC(probs.cv, labs.cv, fn, colScheme=colScheme)
		}
	}
	res <- list(
		fit=predTest.folds,
		pred=predTab,
		prob=probArr
	)
	return(res)
}

# export tables from a trained predictor as TSV. Exported tables include
# - predicted probabilities
# - list of indices for signature features
exportPredictorTables <- function(predRes, prefix){
	for (cl in names(predRes)){
		pr <- predRes[[cl]]
		fn <- paste0(prefix, "_", cl, "_classProbs.tsv")
		tt <- data.frame(sample=rownames(pr$prob), pr$prob, stringsAsFactors=FALSE)
		write.table(tt, fn, sep="\t", row.names=FALSE, col.names=TRUE, quote=FALSE, na="")
		fn <- paste0(prefix, "_", cl, "_signatureIndices.tsv")
		write.table(which(pr$fit$is.imp), fn, sep="\t", row.names=FALSE, col.names=FALSE, quote=FALSE, na="")
	}
}
